#include "AIDriver.h"
#include <cstdio>
#include <iostream>
#include <cmath>


AI_Driver::AI_Driver()
{
	m_turn_left = false;
	m_turn_right = false;
	m_thrusters_on = false;
	m_current_path_node = NULL;
	m_flying_state = FLYING_STATE_GOAL_SEARCHING;
}

AI_Driver::~AI_Driver()
{
	delete_path_nodes();
}

void AI_Driver::update_data(float distance_to_goal, float lander_angle, float sensors[8], float goal_x, float goal_y, 
						    float lander_x, float lander_y, float lander_x_vel, float lander_y_vel, float fuel_left)
{
	m_distance_to_goal = distance_to_goal;
	m_lander_angle = lander_angle;
	memcpy(m_sensor_data, sensors, sizeof(float) * 8);
	m_goal_position = Vector2(goal_x, goal_y);
	m_position = Vector2(lander_x, lander_y);
	m_velocity = Vector2(lander_x_vel, lander_y_vel);
	m_fuel_left = fuel_left;

	m_angle_to_goal = Utils::calc_angle(m_position, m_goal_position);
}

void AI_Driver::update(float dT)
{
	m_turn_left = false;
	m_turn_right = false;
	m_thrusters_on = false;

	// Screwed...
	if(m_fuel_left <= 0)
		m_flying_state = FLYING_STATE_LANDED;

	handle_path_finding(dT);
	handle_flying(dT);
}

void AI_Driver::output_debug_data()
{
	Lunar_AI::Draw_Text( 50.0f, 50.0f, 1.0f, true, GREEN, "X:%f Y%f", m_position.X, m_position.Y );

	// Draw the path nodes
	for(int i = 0; i < m_path_nodes.size(); ++i)
	{
		m_path_nodes[i]->draw();
	}

	// Draw path finding state
	if(m_current_path_node && !m_back_tracking && m_flying_state == FLYING_STATE_GOAL_SEARCHING)
	{
		Utils::draw_line(m_position, m_current_path_node->get_center(), GREEN);
		Lunar_AI::Draw_Text( m_position.X - 1.5, m_position.Y + 1.5f, 0.5f, false, RED, "Goal Searching!");
	}
	else if(m_flying_state == FLYING_STATE_LANDING)
	{
		Lunar_AI::Draw_Text( m_position.X - 1.5, m_position.Y + 1.5f, 0.5f, false, RED, "Landing!");
	}
	else if(m_flying_state == FLYING_STATE_LANDED)
	{
		Lunar_AI::Draw_Text( m_position.X - 1.5, m_position.Y + 1.5f, 0.5f, false, RED, "Landed!");
	}
	else if(m_flying_state == FLYING_STATE_BACK_TRACKING)
	{
		Lunar_AI::Draw_Text( m_position.X - 1.5, m_position.Y + 1.5f, 0.5f, false, RED, "Back Tracking!");
	}
	else if(m_flying_state == FLYING_STATE_RETURN_TO_NODE)
	{
		Lunar_AI::Draw_Text( m_position.X - 1.5, m_position.Y + 1.5f, 0.5f, false, RED, "Retrying!");
	}
	else
	{
		Lunar_AI::Draw_Text( m_position.X - 0.75f, m_position.Y + 1.5f, 0.5f, false, RED, "I'm lost...");
	}

	Lunar_AI::Draw_Text( m_target_location.X, m_target_location.Y - 0.5f, 0.5f, false, RED, "o");
}

void AI_Driver::delete_path_nodes()
{
	for(int i = 0; i < m_path_nodes.size(); ++i)
	{
		delete m_path_nodes[i];
	}

	m_path_nodes.clear();
	m_current_path_node = NULL;
}

float AI_Driver::get_closest_sensor_distance()
{
	float dist = m_sensor_data[0];

	for(int i = 1; i < 8; ++i)
	{
		if(m_sensor_data[i] < dist)
			dist = m_sensor_data[i];
	}

	return dist;
}

bool AI_Driver::is_point_within_path_nodes(Vector2 loc)
{
	for(int i = 0; i < m_path_nodes.size(); ++i)
	{
		Path_Node* node = m_path_nodes[i];

		if(node->is_within_node(loc))
		{ 
			return true;
		}
	}

	return false;
}

bool AI_Driver::is_point_close_to_any_node(Vector2 loc)
{
	for(int i = 0; i < m_path_nodes.size(); ++i)
	{
		Path_Node* node = m_path_nodes[i];

		float distance = node->get_radius() * 1.7f;

		if(Utils::squared_distance_between(loc, node->get_center()) < distance * distance)
		{ 
			return true;
		}
	}

	return false;
}

void AI_Driver::handle_path_finding(float dT)
{
	if(m_flying_state == FLYING_STATE_LANDED ||
	   m_flying_state == FLYING_STATE_LANDING)
	{
		return;
	}

	const float MAX_NODE_RADIUS_SIZE = 5.15f;

	//If we don't have a path node, place one where we are.
	if(!m_current_path_node)
	{
		m_path_nodes.push_back(new Path_Node(m_position, Utils::min(MAX_NODE_RADIUS_SIZE, get_closest_sensor_distance()), m_goal_position, m_angle_to_goal, m_sensor_data, NULL));
		m_current_path_node = m_path_nodes[m_path_nodes.size() - 1];
		m_current_path_node->calculate_valid_paths();
		m_current_path_node->calculate_path_to_pursue();
		m_target_location = m_current_path_node->get_suggested_next_waypoint();
		m_back_tracking = false;
		m_time_taken_to_travel_to_next_waypoint = 0;
	} 
	else if(m_flying_state == FLYING_STATE_GOAL_SEARCHING ||
			m_flying_state == FLYING_STATE_LANDING)
	{
		const float MIN_DISTANCE_TO_REACH_TARGET = 2.5f;

		// If we are close to our target location, spawn a new node.
		if(m_flying_state == FLYING_STATE_GOAL_SEARCHING && 
		   Utils::squared_distance_between(m_position, m_target_location) <= MIN_DISTANCE_TO_REACH_TARGET * MIN_DISTANCE_TO_REACH_TARGET)
		{
			m_path_nodes.push_back(new Path_Node(m_position, MAX_NODE_RADIUS_SIZE, m_goal_position, m_angle_to_goal, m_sensor_data, m_current_path_node));
			m_current_path_node = m_path_nodes[m_path_nodes.size() - 1];
			m_current_path_node->calculate_valid_paths();
			
			calculate_next_valid_path_on_node(m_current_path_node);

			m_time_taken_to_travel_to_next_waypoint = 0;
		}
	}
	else if(m_flying_state == FLYING_STATE_RETURN_TO_NODE)
	{
		const float MIN_DISTANCE_TO_REACH_TARGET = 2.5f;

		// If we are close to our target location, spawn a new node.
		if(Utils::squared_distance_between(m_position, m_target_location) <= MIN_DISTANCE_TO_REACH_TARGET * MIN_DISTANCE_TO_REACH_TARGET)
		{
			calculate_next_valid_path_on_node(m_current_path_node);

			m_flying_state = FLYING_STATE_GOAL_SEARCHING;
			m_time_taken_to_travel_to_next_waypoint = 0;
		}
	}
	else // Back tracking
	{
		// See if we have made it back to our previous node.
		if(m_current_path_node->is_within_node(m_position))
		{
			// Find the next location to attempt from this node.
			m_current_path_node->invalidate_current_path();
			calculate_next_valid_path_on_node(m_current_path_node);
			m_flying_state = FLYING_STATE_GOAL_SEARCHING;
			m_time_taken_to_travel_to_next_waypoint = 0;
		}
	}

	if(m_current_path_node)
	{
		// If our current node has no potential paths, we need to back track.
		if(m_current_path_node->get_num_potential_paths() == 0)
		{
			//Back track
			m_flying_state = FLYING_STATE_BACK_TRACKING;
			m_current_path_node = m_current_path_node->get_node_spawned_from();
			if(m_current_path_node)
			{
				m_target_location = m_current_path_node->get_center();
				m_time_taken_to_travel_to_next_waypoint = 0;
			}
			else
			{
				delete_path_nodes();
				m_flying_state = FLYING_STATE_GOAL_SEARCHING;
			}
		}
	}

	if(m_current_path_node)
	{
		if(m_flying_state != FLYING_STATE_LANDING)
		{
			m_time_taken_to_travel_to_next_waypoint += dT;

			float time_until_retry = 5.5f;

			if(m_current_path_node->is_evaluating_long_path())
			{
				time_until_retry = 7.5f;
			}

			if(m_time_taken_to_travel_to_next_waypoint > time_until_retry)
			{
				if(m_flying_state == FLYING_STATE_RETURN_TO_NODE || m_flying_state == FLYING_STATE_BACK_TRACKING)
				{
					// We have taken too much time to get back to the node, 
					// just dump everything and start from scratch.
					m_flying_state = FLYING_STATE_GOAL_SEARCHING;
					delete_path_nodes();
					m_time_taken_to_travel_to_next_waypoint = 0;
				}
				else
				{
					// We have taken too much time to get to the next waypoint, 
					// go back to the node and pick a different path.
					m_flying_state = FLYING_STATE_RETURN_TO_NODE;
					m_target_location = m_current_path_node->get_center();
					m_time_taken_to_travel_to_next_waypoint = 0;
				}
			}
		}
	}

	m_angle_to_target = Utils::calc_angle(m_position, m_target_location);
}

void AI_Driver::calculate_next_valid_path_on_node(Path_Node* node)
{
	// Check all directions from this new node seeing what is potentially valid.
	// Once we find a suggested path that is not invalid, take that as our next path.
	do
	{
		node->calculate_path_to_pursue();
		m_target_location = node->get_suggested_next_waypoint();

		if(is_point_close_to_any_node(m_target_location))
		{
			node->invalidate_current_path();
		}
		else
		{
			// We found a path that is not invalid, take it.
			break;
		}

	}while(node->get_num_potential_paths() > 0);
}

void AI_Driver::handle_flying(float dT)
{
	while(m_lander_angle < 0)
		m_lander_angle += PI * 2;
	while(m_lander_angle > PI * 2)
		m_lander_angle -= PI * 2;

	switch(m_flying_state)
	{
		case FLYING_STATE_GOAL_SEARCHING:
		case FLYING_STATE_RETURN_TO_NODE:
		case FLYING_STATE_BACK_TRACKING:
		{
			handle_goal_search_flying(dT);
		
			if(m_distance_to_goal < 14.25f && m_position.Y > m_goal_position.Y)
			{
				m_target_location = m_goal_position;
				m_flying_state = FLYING_STATE_LANDING;
			}
		}
		break;

		case FLYING_STATE_LANDING:
		{
			handle_goal_search_flying(dT);
			handle_landing(dT);
		}
		break;

		case FLYING_STATE_LANDED:
		{
			float desired_angle = PI / 2.0f;

			float angle_difference = abs(m_lander_angle - desired_angle);

			if((m_lander_angle - desired_angle > 0 && m_lander_angle - desired_angle < PI) ||
			   (m_lander_angle + 2 * PI - desired_angle > 0 && m_lander_angle + 2 * PI - desired_angle < PI))
			{
				// Turn clock wise / right
				m_turn_right = true;
			}
			else if(angle_difference > 0.00025f)
			{
				// Turn left
				m_turn_left = true;
			}
		}
		break;
	}

	m_previous_position = m_position;
}

void AI_Driver::handle_goal_search_flying_old(float dT)
{
	bool target_location_is_left = m_target_location.X < m_position.X;
	bool target_location_is_right = m_target_location.X > m_position.X;
	bool target_location_is_up = m_target_location.Y > m_position.Y;
	bool target_location_is_down = m_target_location.Y < m_position.Y;

	float x_diff = abs(m_target_location.X - m_position.X);
	float y_diff = abs(m_target_location.Y - m_position.Y);

	// Target is quite a bit to the left.
	if(target_location_is_left && x_diff > 2.0f)
	{
		if(m_velocity.X > -4.0f)
		{
			if(target_location_is_up &&  m_velocity.Y > 2.0f ||
			   target_location_is_down && m_velocity.Y > -1.0f &&
			   m_flying_state != FLYING_STATE_LANDING)
			{
				m_target_angle = 10.0f * PI / 11.0f;
			}
			else
			{
				m_target_angle = 3 * PI / 4.0f;
			}
		}
		else
		{
			m_target_angle = PI / 2.0f;
		}
	}
	// Target is quite a bit to the right
	else if(target_location_is_right && x_diff > 2.0f)
	{
		if(m_velocity.X < 4.0f)
		{
			if(target_location_is_up &&  m_velocity.Y > 2.0f ||
			   target_location_is_down && m_velocity.Y > -1.0f &&
			   m_flying_state != FLYING_STATE_LANDING)
			{
				m_target_angle = PI / 11.0f;
			}
			else
			{
				m_target_angle = PI / 4.0f;
			}
		}
		else
		{
			m_target_angle = PI / 2.0f;
		}
	}
	// It's straight up.
	else
	{
		m_target_angle = PI / 2.0f;
	}


	// If our target angle is counter clock wise.
	if(m_lander_angle - m_target_angle > 0)
	{
		m_turn_right = true;
	}
	else if(m_lander_angle - m_target_angle < 0)
	{
		m_turn_left = true;
	}

	// We are within some threshold of the target angle.
	if(abs(m_lander_angle - m_target_angle ) < PI / 16.0f)
	{
		//Lander is tilted left
		if(m_lander_angle > PI / 2.0f && m_lander_angle < 3 * PI / 2.0f)
		{
			if(m_velocity.X > -4.5f)
			{
				if(x_diff < 2.0f && m_velocity.X < -0.5f)
				{

				}
				else
				{
					m_thrusters_on = true;
				}
			}
		}
		//Lander tilted right
		else
		{
			if(m_velocity.X < 4.5f)
			{
				if(x_diff < 2.0f && m_velocity.X > 0.5f)
				{
						
				}
				else
				{
					m_thrusters_on = true;
				}
			}
		}
	}
		
	if(target_location_is_up)
	{
		if(m_velocity.Y > 4.0f || m_velocity.Y > 0 && y_diff < 2.0f)
		{
			m_thrusters_on = false;
		}
		else if(m_velocity.Y < 1.75f)
		{
			m_thrusters_on = true;
		}
	}
	// If target is down and velocity isn't too fast downward and we don't need to correct our x velocity, ensure the thrusters are off.
	else if(target_location_is_down)
	{
		if(m_velocity.Y > -5 && !((target_location_is_left && m_velocity.X > -1.0 || target_location_is_right && m_velocity.X < 1.0) && m_target_angle != PI / 2.0f))
		{
			m_thrusters_on = false;
		}
		else
		{
			m_thrusters_on = true;
		}
	}
	
	// Safety Overrides
	if(NORTH_SENSOR < 0.4f)
	{
		m_thrusters_on = false;
	}
	else if(SOUTH_SENSOR < 0.4f)
	{
		m_thrusters_on = true;
	}
}

void AI_Driver::handle_goal_search_flying(float dT)
{
	float velocity_magnitude = Utils::magnitude(m_velocity);
	float distance_to_next_waypoint = Utils::distance_between(m_position, m_target_location);

	Vector2 desired_velocity_when_reaching_waypoint = Vector2(0.75 * cos(m_angle_to_target), 0.75 * sin(m_angle_to_target));

	Vector2 estimated_desired_velocity = Vector2(m_target_location.X - (m_position.X + m_velocity.X), 
												 m_target_location.Y + (-GRAVITY * distance_to_next_waypoint / 10.0f) - (m_position.Y + m_velocity.Y * 1.25f));

	if(m_flying_state != FLYING_STATE_LANDING)
	{
		estimated_desired_velocity.X += desired_velocity_when_reaching_waypoint.X;
		estimated_desired_velocity.Y += desired_velocity_when_reaching_waypoint.Y;
	}

	float estimated_desired_angle = Utils::calc_angle(Vector2(0, 0), estimated_desired_velocity);
	float estimated_desired_velocity_magnitude = Utils::magnitude(estimated_desired_velocity);

	float angle_difference = abs(m_lander_angle - estimated_desired_angle);

	if((m_lander_angle - estimated_desired_angle > 0 && m_lander_angle - estimated_desired_angle < PI) ||
	   (m_lander_angle + 2 * PI - estimated_desired_angle > 0 && m_lander_angle + 2 * PI - estimated_desired_angle < PI))
	{
		// Turn clock wise / right
		m_turn_right = true;
	}
	else if(angle_difference > 0.0001f)
	{
		// Turn left
		m_turn_left = true;
	}

	if(abs(m_lander_angle - estimated_desired_angle) < PI / 128.0f || abs(m_lander_angle + 2 * PI - estimated_desired_angle) < PI / 128.0f)
	{
		//if(velocity_magnitude > 3 || estimated_desired_velocity_magnitude > 8)
			m_thrusters_on = true;
	}

	// Safety Overrides
	if(NORTH_SENSOR < 0.4f)
	{
		m_thrusters_on = false;
	}
	else if(SOUTH_SENSOR < 0.4f)
	{
		m_thrusters_on = true;
	}
}

void AI_Driver::handle_landing(float dT)
{
	if(m_distance_to_goal < 1.5f && m_velocity.Y < 0.5f && m_velocity.Y > -0.85f && m_velocity.X < 0.85f && m_velocity.X > -0.85f)
	{
		m_flying_state = FLYING_STATE_LANDED;
	}
}

bool AI_Driver::turning_left()
{
	return m_turn_left;
}

bool AI_Driver::turning_right()
{
	return m_turn_right;
}

bool AI_Driver::thrusters_on()
{
	return m_thrusters_on;
}

