#include "PathNode.h"
#include <iostream>

Path_Node::Path_Node()
{
	memset(m_potential_short_paths, false, sizeof(bool) * 8);
	memset(m_potential_long_paths, false, sizeof(bool) * 8);
	m_node_spawned_from = NULL;
	m_current_path = 0;
	m_radius = 0;
	m_evaluating_long_path = true;
	m_potiential_short_paths_count = 0;
	m_potiential_long_paths_count = 0;
}

Path_Node::Path_Node(Vector2 center, float radius, Vector2 goal_position, float angle_to_goal, float sensor_data[8], Path_Node* spawned_from)
{
	m_center = center;
	m_radius = radius;
	m_goal_position = goal_position;
	m_angle_to_goal = angle_to_goal;
	m_node_spawned_from = spawned_from;
	memcpy(m_sensor_data, sensor_data, sizeof(float) * 8);
	memset(m_potential_short_paths, false, sizeof(bool) * 8);
	memset(m_potential_long_paths, false, sizeof(bool) * 8);
	m_evaluating_long_path = true;
	m_potiential_short_paths_count = MAX_POTENTIAL_PATHS;
	m_potiential_long_paths_count = MAX_POTENTIAL_PATHS;
}

void Path_Node::calculate_valid_paths()
{
	const float SHORT_RADIUS_MULTIPLIER = 2.75f;
	const float LONG_RADIUS_MULTIPLIER = 5.5f;

	for(int i = 0; i < 8; ++i)
	{
		if(m_sensor_data[i] > m_radius * SHORT_RADIUS_MULTIPLIER)
		{
			m_potential_short_paths[i] = true;
			m_potiential_short_paths_count++;
		}

		/*if(m_sensor_data[i] > m_radius * LONG_RADIUS_MULTIPLIER)
		{
			m_potential_long_paths[i] = true;
			m_potiential_long_paths_count++;
		}*/
	}
}

void Path_Node::calculate_path_to_pursue()
{
	int best_path = -1; 
	float smallest_angle_difference = 4 * PI;

	invalidate_path_arrived_from();

	for(int i = 0; i < 8; ++i)
	{
		if(m_potential_short_paths[i] || m_potential_long_paths[i])
		{
			float angle_difference = Utils::min(abs(m_angle_to_goal - sensor_angles[i]), abs(m_angle_to_goal - (sensor_angles[i] + 2 * PI))); 

			if(angle_difference < smallest_angle_difference)
			{
				smallest_angle_difference = angle_difference;
				best_path = i;
			}
		}
	}

	if(best_path != -1)
	{
		m_current_path = best_path;

		float short_distance_to_attempt = m_radius * 2.0f;
		float long_distance_to_attempt = m_radius * 4.75f;

		Vector2 short_waypoint_position = Vector2(m_center.X + short_distance_to_attempt * cos_of_sensor_angles[m_current_path], m_center.Y + short_distance_to_attempt * sin_of_sensor_angles[m_current_path]);
		Vector2 long_waypoint_position = Vector2(m_center.X + long_distance_to_attempt * cos_of_sensor_angles[m_current_path], m_center.Y + long_distance_to_attempt * sin_of_sensor_angles[m_current_path]);

		float distance_from_goal_to_short_waypoint = Utils::squared_distance_between(m_goal_position, short_waypoint_position);
		float distance_from_goal_to_long_waypoint = Utils::squared_distance_between(m_goal_position, long_waypoint_position);

		if(m_potential_long_paths[best_path] && distance_from_goal_to_short_waypoint >= distance_from_goal_to_long_waypoint)
		{
			m_next_suggested_waypoint = long_waypoint_position;
			m_evaluating_long_path = true;
		}
		else
		{
			m_next_suggested_waypoint = short_waypoint_position;
			m_evaluating_long_path = false;
			m_potential_long_paths[best_path] = false;
		}
	}
}

void Path_Node::invalidate_current_path()
{
	/*if(m_attempting_long_path)
	{
		m_attempting_long_path = false;
		return;
	}*/

	if(m_evaluating_long_path)
	{
		m_potiential_long_paths_count--;
		m_potential_long_paths[m_current_path] = false;
		m_evaluating_long_path = false;
	}
	else
	{
		m_potiential_short_paths_count--;
		m_potential_short_paths[m_current_path] = false;
		m_evaluating_long_path = false;
	}
}

float Path_Node::get_radius()
{
	return m_radius;
}

Vector2 Path_Node::get_center()
{
	return m_center;
}

bool Path_Node::is_within_node(Vector2 pos)
{
	return Utils::squared_distance_between(pos, m_center) <= m_radius * m_radius;
}

int Path_Node::get_num_potential_paths()
{
	return m_potiential_long_paths_count + m_potiential_short_paths_count;
}

Path_Node* Path_Node::get_node_spawned_from()
{
	return m_node_spawned_from;
}

Vector2 Path_Node::get_suggested_next_waypoint()
{
	return m_next_suggested_waypoint;
}

bool Path_Node::is_evaluating_long_path()
{
	return m_evaluating_long_path;
}

void Path_Node::draw()
{
	bool available_path = false;

	for(int i = 0; i < MAX_POTENTIAL_PATHS; ++i)
	{
		if(m_potential_long_paths[i] || m_potential_short_paths[i])
		{
			available_path = true;
			break;
		}
	}

	if(available_path)
		Utils::draw_circle(m_center, m_radius, GREEN);
	else
		Utils::draw_circle(m_center, m_radius, RED);

	if(m_node_spawned_from)
		Utils::draw_line(m_center, m_node_spawned_from->m_center, GREEN);

	// Draw potential paths

	for(int i = 0; i < MAX_POTENTIAL_PATHS; ++i)
	{
		if(m_potential_long_paths[i] || m_potential_short_paths[i])
		{
			Utils::draw_line(m_center, Vector2(m_center.X + m_radius * cos_of_sensor_angles[i], m_center.Y + m_radius * sin_of_sensor_angles[i]), TRANSPARENT_BLUE);
		}
	}

	if(m_potiential_short_paths_count > 0 || m_potiential_long_paths_count > 0)
	{
		Utils::draw_line(m_center,  Vector2(m_center.X + m_radius * cos_of_sensor_angles[m_current_path], m_center.Y + m_radius * sin_of_sensor_angles[m_current_path]), ORANGE);
	}

	/*if(m_potential_paths[4])
		draw_line(m_center, Vector2(m_center.X + m_radius * COS_0, m_center.Y + m_radius * SIN_0), TRANSPARENT_BLUE);
	if(m_potential_paths[2])
		draw_line(m_center, Vector2(m_center.X + m_radius * COS_PI_4, m_center.Y + m_radius * SIN_PI_4), TRANSPARENT_BLUE);
	if(m_potential_paths[1])
		draw_line(m_center, Vector2(m_center.X + m_radius * COS_PI_2, m_center.Y + m_radius * SIN_PI_2), TRANSPARENT_BLUE);
	if(m_potential_paths[0])
		draw_line(m_center, Vector2(m_center.X + m_radius * COS_3_PI_4, m_center.Y + m_radius * SIN_3_PI_4), TRANSPARENT_BLUE);
	if(m_potential_paths[3])
		draw_line(m_center, Vector2(m_center.X + m_radius * COS_PI, m_center.Y + m_radius * SIN_PI), TRANSPARENT_BLUE);
	if(m_potential_paths[5])
		draw_line(m_center, Vector2(m_center.X + m_radius * COS_5_PI_4, m_center.Y + m_radius * SIN_5_PI_4), TRANSPARENT_BLUE);
	if(m_potential_paths[6])
		draw_line(m_center, Vector2(m_center.X + m_radius * COS_3_PI_2, m_center.Y + m_radius * SIN_3_PI_2), TRANSPARENT_BLUE);
	if(m_potential_paths[7])
		draw_line(m_center, Vector2(m_center.X + m_radius * COS_7_PI_4, m_center.Y + m_radius * SIN_7_PI_4), TRANSPARENT_BLUE);*/
		
}

void Path_Node::invalidate_path_arrived_from()
{
	if(m_node_spawned_from)
	{
		if(m_evaluating_long_path)
		{
			m_potential_long_paths[7 - m_node_spawned_from->m_current_path] = false;
		}
		else
		{
			m_potential_short_paths[7 - m_node_spawned_from->m_current_path] = false;
		}
	}
}